﻿using System;
using System.Threading;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace TVSharp
{
    public unsafe partial class MainForm : Form
    {
        private readonly RtlSdrIO _rtlDevice = new RtlSdrIO();
        private bool _isDecoding;
        private bool _initialized;
        private Bitmap _buffer;
        private byte[] rgbValues;
        private PixelFormat _pxf;
        private Rectangle rect;
        private int _detectLevel;
        private float _detectorLevelCoef;
        private int _x;
        private int _y;
        private int _pictureWidth;
        private int _pictureHeight;
        private int _pixelCounter;
        private int _maxSignalLevel;
        private int _agcSignalLevel;
        private int _blackLevel;
        private int _bright;
        private float _contrast;
        private SettingsMemoryEntry _settings;
        private readonly SettingsPersister _settingsPersister = new SettingsPersister ();
        
        public MainForm()
        {
            InitializeComponent();

            frequencyNumericUpDown_ValueChanged(null, null);
            try
            {
                _rtlDevice.Open();

                var devices = DeviceDisplay.GetActiveDevices();
                deviceComboBox.Items.Clear();
                deviceComboBox.Items.AddRange(devices);

                //_initialized = true;
                deviceComboBox.SelectedIndex = 0;
                deviceComboBox_SelectedIndexChanged(null, null);
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message);
            }
        }

        #region GUI Controls

        private void startBtn_Click(object sender, EventArgs e)
        {
            if (!_isDecoding)
            {
                StartDecoding();
            }
            else
            {
                StopDecoding();
            }
            startBtn.Text = _isDecoding ? "Stop" : "Start";
            deviceComboBox.Enabled = !_rtlDevice.Device.IsStreaming;
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (_isDecoding)
            {
                StopDecoding();
            }

            _settings.Contrast = contrastTrackBar.Value;
            _settings.Brightnes = brightnesTrackBar.Value;
            _settings.Frequency = (long)frequencyNumericUpDown.Value;
            _settings.FrequencyCorrection = (int)frequencyCorrectionNumericUpDown.Value;
            _settings.PictureHeight = _pictureHeight;
            _settings.PictureWidth = _pictureWidth;
            _settings.DetectorLevel = _detectorLevelCoef;

            _settingsPersister.PersistSettings(_settings);
        }

        private void deviceComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            //if (!_initialized)
            //{
            //    return;
            //}
            var deviceDisplay = (DeviceDisplay)deviceComboBox.SelectedItem;
            if (deviceDisplay != null)
            {
                try
                {
                    _rtlDevice.SelectDevice(deviceDisplay.Index);
                    _rtlDevice.Frequency = (uint)frequencyNumericUpDown.Value;
                    _rtlDevice.Device.Samplerate = 2000000;
                    _rtlDevice.Device.UseRtlAGC = false;
                    _rtlDevice.Device.UseTunerAGC = false;
                    _initialized = true;
                }
                catch (Exception ex)
                {
                    deviceComboBox.SelectedIndex = -1;
                    _initialized = false;
                    MessageBox.Show(this, ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                ConfigureDevice();
                ConfigureGUI();
            }
        }

        private void tunerGainTrackBar_Scroll(object sender, EventArgs e)
        {
            if (!_initialized)
            {
                return;
            }
            var gain = _rtlDevice.Device.SupportedGains[tunerGainTrackBar.Value];
            _rtlDevice.Device.TunerGain = gain;
            gainLabel.Text = gain / 10.0 + " dB";
        }

        private void frequencyCorrectionNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            if (!_initialized)
            {
                return;
            }
            _rtlDevice.Device.FrequencyCorrection = (int)frequencyCorrectionNumericUpDown.Value;
        }

        #endregion

        #region Private Methods

        private void ConfigureGUI()
        {
            startBtn.Enabled = _initialized;
            if (!_initialized)
            {
                return;
            }
            _settings = _settingsPersister.ReadSettings();
            tunerTypeLabel.Text = _rtlDevice.Device.TunerType.ToString();
            tunerGainTrackBar.Maximum = _rtlDevice.Device.SupportedGains.Length - 1;
            tunerGainTrackBar.Value = tunerGainTrackBar.Maximum;
            brightnesTrackBar.Value = _settings.Brightnes;
            brightnesTrackBar_Scroll(null, null);
            contrastTrackBar.Value = _settings.Contrast;
            contrastTrackBar_Scroll(null, null);
            frequencyCorrectionNumericUpDown.Value = _settings.FrequencyCorrection;
            frequencyCorrectionNumericUpDown_ValueChanged(null, null);
            frequencyNumericUpDown.Value = _settings.Frequency;
            frequencyNumericUpDown_ValueChanged(null, null);


            for (var i = 0; i < deviceComboBox.Items.Count; i++)
            {
                var deviceDisplay = (DeviceDisplay)deviceComboBox.Items[i];
                if (deviceDisplay.Index == _rtlDevice.Device.Index)
                {
                    deviceComboBox.SelectedIndex = i;
                    break;
                }
            }

            _detectorLevelCoef = _settings.DetectorLevel;//0.77;
            if (_detectorLevelCoef == 0) _detectorLevelCoef = 0.77f;

            _pictureHeight = _settings.PictureHeight;//313;
            if (_pictureHeight == 0) _pictureHeight = 313;

            _pictureWidth = _settings.PictureWidth;//128;
            if (_pictureWidth == 0) _pictureWidth = 128;

            _pxf = PixelFormat.Format24bppRgb;
            _buffer = new Bitmap(_pictureWidth, _pictureHeight, _pxf);
            rect = new Rectangle(0, 0, _buffer.Width, _buffer.Height);
            int lengthOfRGBArray = _buffer.Width * _buffer.Height * 3;
            rgbValues = new byte[lengthOfRGBArray];
            
            SetStyle(ControlStyles.OptimizedDoubleBuffer, true);
            SetStyle(ControlStyles.DoubleBuffer, true);
            SetStyle(ControlStyles.UserPaint, true);
            SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            UpdateStyles();
        }

        private void ConfigureDevice()
        {
            frequencyCorrectionNumericUpDown_ValueChanged(null, null);
            tunerGainTrackBar_Scroll(null, null);
        }

        private void StartDecoding()
        {
            try
            {
                _rtlDevice.Start(rtl_SamplesAvailable);
            }
            catch (Exception e)
            {
                StopDecoding();
                MessageBox.Show("Unable to start RTL device\n" + e.Message);
                return;
            }
            _isDecoding = true;
        }

        private void StopDecoding()
        {
            _rtlDevice.Stop();
            _isDecoding = false;
        }

        #endregion

        #region Samples Callback

        private void rtl_SamplesAvailable(object sender, Complex* buf, int length)
        {
            _maxSignalLevel = 0;
            for (var i = 0; i < 300; i++)
            {
                var real = buf[i].Real;
                var imag = buf[i].Imag;

                var mag = (int)(Math.Sqrt(real * real + imag * imag));

                _maxSignalLevel = Math.Max(mag, _maxSignalLevel);
            }
            _agcSignalLevel = _maxSignalLevel;
            _detectLevel = Convert.ToInt32(_maxSignalLevel * _detectorLevelCoef);
            for (var i = 0; i < length; i++)
            {
                var real = buf[i].Real;
                var imag = buf[i].Imag;

                var mag = (int)(Math.Sqrt(real * real + imag * imag));
                DrawPixel(mag);
            }
        }

        #endregion

        private void fpsTimer_Tick(object sender, EventArgs e)
        {
            OscillocsopePanel.Refresh();
            if (_agcSignalLevel > 130 && _isDecoding)
            {
                if (tunerGainTrackBar.Value > tunerGainTrackBar.Minimum)
                {
                    tunerGainTrackBar.Value = tunerGainTrackBar.Value - 1;
                    tunerGainTrackBar_Scroll(null, null);
                }
            }

            if (_agcSignalLevel < 90 && _isDecoding)
            {
                if (tunerGainTrackBar.Value < tunerGainTrackBar.Maximum)
                {
                    tunerGainTrackBar.Value = tunerGainTrackBar.Value + 1;
                    tunerGainTrackBar_Scroll(null, null);
                }
            }
        }

        private void DrawPixel(int mag)
        {

            if (mag > _detectLevel)
            {
                _pixelCounter++;
            }
            else if (_pixelCounter > 0)
            {

                if (_pixelCounter > 40)
                {
                    var centr = (int)_pictureHeight / 2;
                    if (_y < centr && _y > 0) _y--;
                    if (_y > centr) _y++;
                }
                else if (_pixelCounter > 5 && _pixelCounter < 20 && (_y < 25 || _y > _pictureHeight - 25))
                {
                    var centr = (int)_pictureWidth / 2;
                    if (_x < centr && _x > 0) _x--;
                    if (_x > centr) _x++;
                }
                _pixelCounter = 0;
            }
            if (_x == 4 && _y == 5)
            {
                _blackLevel = Convert.ToInt32(_blackLevel * 0.99 + mag * 0.01);
                mag = 0;
            }
            if (_x >= _pictureWidth)
            {
                _x = 0;
                _y++;
            }

            if (_y >= _pictureHeight)
            {
                _y = 0;
            }

            var pixelAdress = (_y * _pictureWidth + _x) * 3;
            var temp = ((_blackLevel - mag) * _contrast) + _bright;
            if (temp < 0) temp = 0;
            if (temp > 255) temp = 255;
            var brightnes = Convert.ToByte(temp);
            rgbValues[pixelAdress] = brightnes;
            rgbValues[pixelAdress + 1] = brightnes;
            rgbValues[pixelAdress + 2] = brightnes;
            _x++;
            
        }

        private void CopyScreenToBitmap()
        {

            // Lock the bitmap's bits.
            BitmapData bmpData = _buffer.LockBits(rect, ImageLockMode.ReadWrite, _pxf);

            // Get the address of the first line.
            IntPtr ptr = bmpData.Scan0;

            // Copy the RGB values back to the bitmap
            Marshal.Copy(rgbValues, 0, ptr, rgbValues.Length);

            // Unlock the bits.
            _buffer.UnlockBits(bmpData);
        }


        private void OscillocsopePanel_Paint(object sender, PaintEventArgs e)
        {
            CopyScreenToBitmap();
            e.Graphics.DrawImage(_buffer, 0, 0, 400, 300);
        }

        private void frequencyNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            {
                if (!_initialized)
                {
                    return;
                }
                //StopDecoding();
                _rtlDevice.Device.Frequency = (uint)frequencyNumericUpDown.Value;
                //StartDecoding();
            }
        }

        private void brightnesTrackBar_Scroll(object sender, EventArgs e)
        {
            _bright = brightnesTrackBar.Value;
        }

        private void contrastTrackBar_Scroll(object sender, EventArgs e)
        {
            _contrast = (contrastTrackBar.Value) / 10;
        }

    }

    public class DeviceDisplay
    {
        public uint Index { get; private set; }
        public string Name { get; set; }

        public static DeviceDisplay[] GetActiveDevices()
        {
            var count = NativeMethods.rtlsdr_get_device_count();
            var result = new DeviceDisplay[count];

            for (var i = 0u; i < count; i++)
            {
                var name = NativeMethods.rtlsdr_get_device_name(i);
                result[i] = new DeviceDisplay { Index = i, Name = name };
            }

            return result;
        }

        public override string ToString()
        {
            return Name;
        }
    }
}
